from influxdb import InfluxDBClient, SeriesHelper
import os
from datetime import datetime


def _create_client():
    """
    returns a InfluxDBClient object. Always tests connectivity before returning.
    """
    client = InfluxDBClient(
        host=os.getenv("INFLUX_HOST", "localhost"),
        port=int(os.getenv("INFLUX_PORT", 8086)),
        username=os.getenv("INFLUX_USER", ""),
        password=os.getenv("INFLUX_PASSWORD", ""),
        database=os.getenv("INFLUX_DB"),
    )

    # Check if client is available
    try:
        client.ping()
    except Exception:
        raise ConnectionError("Could not connect to InfluxDB")
    return client


def _create_series_helper(custom_tags={}):
    """
    Create a custom InfluxDB Series Helper, dynamically adding custom_tags
    """
    custom_tags_list = list(custom_tags.keys())

    class SensorSeriesHelper(SeriesHelper):
        class Meta:
            client = _create_client()
            series_name = "thermostat"
            fields = ["time", "temperature"]
            tags = [
                "user",
                "home_name",
                "home_id",
                "room_name",
                "room_id",
            ] + custom_tags_list
            bulk_size = 512
            autocommit = True

    return SensorSeriesHelper


def store_measurements(
    user: str,
    home_id: str,
    home_name: str,
    room_id: str,
    room_name: str,
    room_data: dict,
    custom_tags: dict,
):
    """
    Stores measurements Netatmo-data in InfluxDB
    """
    series_helper = _create_series_helper(custom_tags)

    # TODO: Allow for custom_tags to be logically assigned, rather than fixed

    # Loop over room_data and add to series_helper
    for key, value in room_data.items():
        series_helper(
            user=user,
            home_name=home_name,
            home_id=home_id,
            room_name=room_name,
            room_id=room_id,
            time=_parse_timestamp(int(key)),
            temperature=float(
                round(value[0], 1)
            ),  # For influxdb the type should always be float
            **custom_tags  # Unpack custom_tags dict
        )

    # Commit remainder
    series_helper.commit()


def _parse_timestamp(ts: int):
    """
    Creates UTC-timestamp from Epoch ts
    """
    return datetime.utcfromtimestamp(ts).isoformat()



import getpass
import os
from datetime import datetime
from sqlite3 import IntegrityError

import requests
from logger import logger

import db

NETATMO_BASE = os.getenv("NETATMO_BASE")
# Specific to API security
CLIENT_ID = os.getenv("CLIENT_ID")
if not CLIENT_ID:
    raise EnvironmentError("Missing CLIENT_ID")
CLIENT_SECRET = os.getenv("CLIENT_SECRET")
if not CLIENT_SECRET:
    raise EnvironmentError("Missing CLIENT_SECRET")


def get_access_token(username):
    logger.debug(f"Fetching access token for {username}")
    user = db.get_user(username)
    if user is None:  # User does not exist yet, create initial tokens
        logger.debug(f"First time use for user {username} - initializing tokens")
        token = init_tokens(username)
        return token["access_token"]

    # Check if token has expired
    logger.debug(f"Checking if token expired")
    now = datetime.utcnow().timestamp()
    expires = user["expires_ts"]
    if now >= expires:
        # If the token has expired, refresh
        logger.debug(f"Token expired, getting a new token")
        token = refresh_access_token(username)
        return token["access_token"]

    # Everything is OK
    return user["access_token"]


def init_tokens(username):
    logger.debug(f"Requesting password for first time user {username}")
    # Request password in terminal
    password = getpass.getpass(prompt=f"Password for {username}: ")

    # Create request
    url = NETATMO_BASE + "oauth2/token"
    data = {
        "grant_type": "password",
        "client_id": CLIENT_ID,
        "client_secret": CLIENT_SECRET,
        "username": username,
        "password": password,
        "scope": "read_thermostat write_thermostat",
    }
    resp = requests.post(url, data=data)
    if resp.status_code != 200:
        logger.critical(f"Could not get user access tokens for user {username}")
        raise ConnectionError(f"Something went wrong: {resp.json()}")

    # Process data and save to db
    logger.debug(f"Successfully got information for user {username}")
    tkn = resp.json()
    now = datetime.utcnow().timestamp()
    expires_ts = now + tkn["expires_in"]
    try:
        logger.debug(f"Creating database entry for user {username}")
        db.create_user(username, tkn["refresh_token"], tkn["access_token"], expires_ts)
    except IntegrityError:
        logger.warning(
            f"User {username} already exists in database.\
            Only updating access_token."
        )
        db.update_user(username, tkn["access_token"], expires_ts)
    return tkn


def refresh_access_token(username):
    # Get user
    user = db.get_user(username)
    logger.debug(f"Getting refresh token for user {username}")
    refresh_token = user["refresh_token"]

    # Create request
    url = NETATMO_BASE + "oauth2/token"
    data = {
        "grant_type": "refresh_token",
        "refresh_token": refresh_token,
        "client_id": CLIENT_ID,
        "client_secret": CLIENT_SECRET,
    }
    resp = requests.post(url, data=data)
    if resp.status_code != 200:
        logger.critical(
            f"Could not connect to the Netatmo API or\
            Refresh Token is invalid"
        )
        raise ConnectionError("Could not connect to Netatmo API")

    # Process data and save in db
    logger.debug(f"Storing new access token in database")
    tkn = resp.json()
    now = datetime.utcnow().timestamp()
    expires_ts = now + tkn["expires_in"]

    # Update user
    db.update_user(username, tkn["access_token"], expires_ts)
    return tkn



import data
import store
import db
import argparse
from parser import validate_home_arg
from logger import logger


def run(args):
    logger.debug("Starting Netatmo-to-InfluxDB")
    user = args.user

    # Create db if not exist
    db.init_db(clear_db=args.clear_db)

    # Get user home data
    all_homes = data.get_home(user)

    # If --get-home: print home information and exit
    if args.get_home:
        for home_id in all_homes.keys():
            home = all_homes[home_id]
            print(f"\t{home['name']} ({home_id}) has the following rooms:")
            for room in home["rooms"]:
                print(f"\t\t{room['name']} ({room['id']})")
        return

    # Validate home arguments if available
    homes, err = validate_home_arg(args, all_homes)
    if err is not None:
        return

    # Get measurements
    measurements = data.get_measurements(user, homes, args.dry)

    # Prepare custom_tags
    custom_tags = {}
    if args.custom_tags:
        logger.debug("Preparing custom tags")
        try:
            [
                custom_tags.update({tag: value})
                for tag, value in (ct.split(":") for ct in args.custom_tags)
            ]
        except Exception:
            logger.error("Failed unpacking custom tags")

    # Store measurements
    if not args.dry:
        logger.debug("Preparing data for transfer to InfluxDB")
        for home_id, home_data in measurements.items():
            home_name = home_data["name"]
            for room_id, room_values in home_data["rooms"].items():
                room_name = room_values["name"]
                room_data = room_values["data"]
                if len(room_data.keys()) > 0:
                    logger.debug(
                        f"Sending [home] {home_name} [room] {room_name} data to InfluxDB"
                    )
                    store.store_measurements(
                        user,
                        home_id,
                        home_name,
                        room_id,
                        room_name,
                        room_data,
                        custom_tags,
                    )

    else:
        logger.debug("Using --dry, terminating.")


if __name__ == "__main__":
    username = []
    parser = argparse.ArgumentParser(description="Gather thermostat data from Netatmo")

    # User
    parser.add_argument("user", help="User to parse")

    # Arguments
    parser.add_argument(
        "--home",
        action="append",
        nargs="*",
        metavar=("home_id", ["room_id"]),
        help="Homes and rooms to parse.\
            Use format --home {home_id_1} {room_id_1} {room_id2} ...\
                --home {home_id_2} ...",
    )
    parser.add_argument(
        "--custom-tags",
        nargs="*",
        metavar=("tag:value"),
        help="Provide custom tags for InfluxDB.\
            Format: --custom-tags tag:value tag:value",
    )

    # Functional
    parser.add_argument(
        "--get-home", action="store_true", help="Get home and room information"
    )
    parser.add_argument("--all", action="store_true", help="Parse all homes and rooms")
    parser.add_argument(
        "--dry", action="store_true", help="Do a dry-run (don't store in InfluxDB)"
    )
    parser.add_argument(
        "--clear-db",
        action="store_true",
        help="Wipes database from users and import history",
    )
    parser.add_argument(
        "--interactive",
        action="store_true",
        help="Allows interactive use (ignores all other args)",
    )

    args = parser.parse_args()
    if not args.interactive:
        run(args)



from logger import logger


def validate_home_arg(args, homes):
    """
    Validates and parses supplied home details against all known home- and room-objects.
    """
    logger.debug("Validating home args")
    user = args.user
    if args.home and args.all:
        logger.critical("Only use one of the following arguments: --all OR --home ...")
        return None, True

    if not (args.home or args.all):
        logger.critical("Use one of --all or --home ...")
        return None, True

    if args.home:
        arg_homes = {}
        for home in args.home:
            # Home ID provided in the argument
            if len(home) == 0:
                logger.critical(f"No data provided for --home")
                return None, True
            arg_home_id = home[0]
            # Home IDs that are found for the user
            home_ids = homes.keys()
            # Verify if home ID is valid
            if arg_home_id not in home_ids:
                logger.critical(f"Home {arg_home_id} not found for user {user}")
                return None, True

            arg_homes[arg_home_id] = {"name": homes[arg_home_id]["name"], "rooms": []}

            # Room IDs provided in argument
            arg_room_ids = home[1:]
            # Room IDs that are found for the user
            room_ids = [room["id"] for room in homes[arg_home_id]["rooms"]]
            # Verify if room ID is at least a subset of the available room IDs
            if not all(x in room_ids for x in arg_room_ids):
                logger.critical(f"Invalid rooms supplied for user {user}")
                return False

            if len(arg_room_ids) == 0:
                logger.critical(f"No room data provided for --home {arg_home_id}")
                return None, True

            for room in room_ids:
                room_names = [
                    {room["id"]: room["name"]} for room in homes[arg_home_id]["rooms"]
                ]
                arg_homes[arg_home_id]["rooms"].append(
                    {"id": room, "name": room_names[0][room]}
                )
        return arg_homes, None
    logger.debug("Using all homes and rooms")
    return homes, None



import json
import os

from logger import logger
import requests

import db
import tokens

from datetime import datetime

NETATMO_BASE = os.getenv("NETATMO_BASE")


def get_home(username):
    """
    Get all home and room info. Returns a dict with the following strucuture

    ```
    {
     "home_id_1":
        {"name": str,
         "rooms": [
             {"id": str,
              "name": str
             }, ...
         ]
        },
     "home_id_2": ...
    }
    ```
    """
    logger.debug(f"Retrieving home information for user {username}")
    url = NETATMO_BASE + "api/homesdata"
    params = {"access_token": tokens.get_access_token(username)}
    resp = requests.get(url, params=params)

    if resp.status_code != 200:
        logger.critical(f"Retrieving home information for user {username} failed")
        raise ConnectionError(f"Failed getting data: {resp.content}")

    parsed_resp = json.loads(resp.content)
    homes = {}
    logger.debug("Parsing home information")
    for home in parsed_resp["body"]["homes"]:
        homes[home["id"]] = {"name": home["name"], "rooms": []}
        for room in home["rooms"]:
            homes[home["id"]]["rooms"].append({"id": room["id"], "name": room["name"]})
    return homes


def get_measurements(username, homes, dry):
    """
    Retrieve thermostat measurements for all homes and rooms
    """
    logger.debug(f"Starting retrieval of measurements for user {username}")
    measurements = {}
    for home_id in homes.keys():  # Every home
        home = homes[home_id]
        home_name = home["name"]
        logger.debug(f"Retrieving data for home {home_name} ({home_id})")
        measurements[home_id] = {"id": home_id, "name": home_name, "rooms": {}}
        for room in home["rooms"]:  # Every room in home
            room_id = room["id"]
            room_name = room["name"]
            logger.debug(
                f"Preparing retrieval of data for room {room_name} ({room_id})"
            )
            measurements[home_id]["rooms"][room_id] = {
                "id": room_id,
                "name": room_name,
                "data": {},
            }
            # Get last requested object
            last_record = db.get_last_record(username, home_id, room_id)
            # Add 30 minutes to that last record or use None as start date
            start_date = last_record["end_ts"] + 1800 if last_record else None

            while True:  # Keep requesting data until last page is reached
                # Parse date for logger
                parsed_start_date = (
                    "scratch"
                    if start_date is None
                    else datetime.utcfromtimestamp(int(start_date)).isoformat()
                )
                logger.debug(
                    f"Retrieving data for room {room_name} ({room_id}) from {parsed_start_date}"  # noqa: E501
                )

                # Start collecting data
                collection = collect(
                    start_date, username, home_id=home_id, room_id=room_id, dry=dry
                )
                if collection:
                    measurements[home_id]["rooms"][room_id]["data"].update(
                        collection["data"]
                    )
                    start_date = (
                        collection["end_ts"] + 1800
                    )  # Add thirty minute after last record
                    count = collection["count"]
                    logger.debug(
                        f"Retrieved {count} records for room {room_name} ({room_id})"
                    )
                    if count < 1024:  # Last page
                        break
                else:  # No data
                    logger.debug(f"No data for for room {room_name} ({room_id})")
                    break

    return measurements


def collect(start_date, username, home_id, room_id, dry):
    url = NETATMO_BASE + "api/getroommeasure"
    params = {
        "access_token": tokens.get_access_token(username),
        "home_id": home_id,
        "room_id": room_id,
        "scale": "30min",
        "type": "temperature",
        "date_begin": str(start_date),
        "real_time": True,
        "optimize": False,
    }
    resp = requests.get(url, params=params)

    if resp.status_code != 200:
        raise ConnectionError(f"Failed getting data: {resp.content}")

    parsed_resp = json.loads(resp.content)
    data = parsed_resp["body"]
    count = len(parsed_resp["body"])
    if count > 0 and not dry:
        start_ts = int(min(parsed_resp["body"].keys()))
        end_ts = int(max(parsed_resp["body"].keys()))
        db.create_record(username, home_id, room_id, start_ts, end_ts, count)
        return {"data": data, "start_ts": start_ts, "end_ts": end_ts, "count": count}

    return None



import sqlite3
from datetime import datetime

DB_FILE = "netatmo.db"


def init_db(clear_db=False):
    """
    Initialize database. The clear_db argument allows for a clean start.
    """
    if clear_db:
        sql_clear_users = """ DROP TABLE IF EXISTS users; """
        sql_clear_records = """ DROP TABLE IF EXISTS records; """
        _execute(sql_clear_users)
        _execute(sql_clear_records)
    sql_create_users_table = """ CREATE TABLE IF NOT EXISTS users (
                            username text PRIMARY KEY,
                            refresh_token text,
                            access_token text,
                            expires_ts int
                        ); """
    sql_create_records_table = """ CREATE TABLE IF NOT EXISTS records (
                            username text,
                            home_id text,
                            room_id text,
                            start_ts int,
                            end_ts int,
                            count int,
                            created_ts int
                        );"""
    _execute(sql_create_users_table)
    _execute(sql_create_records_table)


def _create_connection():
    """
    Create a sqlite3 connection object
    """
    try:
        conn = sqlite3.connect(DB_FILE)
        return conn
    except Exception as e:
        print(e)

    return None


def _execute(sql: str):
    """
    Execute SQL
    """
    conn = _create_connection()
    try:
        c = conn.cursor()
        c.execute(sql)
        conn.commit()
        conn.close()
    except Exception as e:
        print(e)


def _dict_factory(cursor, row):
    """
    Creates a dict with row-names as keys for every query.
    """
    d = {}
    for idx, col in enumerate(cursor.description):
        d[col[0]] = row[idx]
    return d


def create_user(username: str, refresh_token: str, access_token: str, expires_ts: int):
    """
    Create a user
    """
    conn = _create_connection()
    sql = f""" INSERT INTO users(username, refresh_token, access_token, expires_ts)
              VALUES(\'{username}\',
                    \'{refresh_token}\',
                    \'{access_token}\',
                    {expires_ts}) """
    cur = conn.cursor()
    cur.execute(sql)
    conn.commit()
    conn.close()


def update_user(username: str, access_token: str, expires_ts: int):
    """
    Update a user access token and expiration timestamp.
    """
    conn = _create_connection()
    sql = f""" UPDATE users SET
                access_token = '{access_token}',
                expires_ts = {expires_ts}
                WHERE username = '{username}'"""
    cur = conn.cursor()
    cur.execute(sql)
    conn.commit()
    conn.close()


def get_user(username: str):
    """
    Retrieve a user by username (username, refresh_token, access_token, expires_ts)
    """
    conn = _create_connection()
    conn.row_factory = _dict_factory
    cur = conn.cursor()
    cur.execute(f"SELECT * FROM users WHERE username='{username}'")
    user = cur.fetchone()
    return user


def create_record(
    username: str, home_id: str, room_id: str, start_ts: int, end_ts: int, count: int
):
    """
    Create a record of import query
    """
    created_ts = datetime.utcnow().timestamp()
    conn = _create_connection()
    sql = f""" INSERT INTO records(username, home_id, room_id,
                                   start_ts, end_ts, count, created_ts)
            VALUES(
                \'{username}\',
                \'{home_id}\',
                \'{room_id}\',
                {start_ts},
                {end_ts},
                {count},
                {created_ts}) """
    cur = conn.cursor()
    cur.execute(sql)
    conn.commit()
    conn.close()


def get_last_record(username, home_id, room_id):
    """
    Get last record by username, home, and room.
    """
    conn = _create_connection()
    conn.row_factory = _dict_factory
    cur = conn.cursor()
    sql = f""" SELECT * FROM records
            WHERE username = \'{username}\' AND
                home_id=\'{home_id}\' AND
                room_id=\'{room_id}\' AND
                created_ts=(SELECT MAX(created_ts)
            FROM records WHERE username=\'{username}\' AND
                            home_id=\'{home_id}\' AND
                            room_id=\'{room_id}\')"""
    cur.execute(sql)
    record = cur.fetchone()
    return record
